//
// Actor.cs - Defines a generic set of interfaces for the objects used in the game.
//            Logical game objects (such as the player, trees, rocks, etc) subclass
//            this object.
// 
// Each Actor has an "Image Size", and "Object Size", and a "Collision Size"
//   These are all read-only properties; their values are defined by the subclass.
//     Collision Size - The 3D dimensions to use when calculating collisions.
//       The X, Y, and Z values for collision will be calculated WRT the object's
//       center.
//     Object Size - The 3D dimensions of the object.
//     Image Size - Might not be applicable any more, but this stores the image size
//       of this object.
//
// Each Actor has a generic "state" variable.  This state variable can be used to track
//   any kind of state for the Actor.  (For the player, we might have "ALIVE" and "DEAD".
//   Valid "state" values are not given here; it is up to the subclass to declare
//     constants that represent the valid states of the object.
//
// Drawing/animating an Actor object is done by calling the animate() method and
// the draw() method, where the destination rectangle is specified.  animate() 
// tells the object to adjust its animation frame, when it is appropriate to do so. 
//
// animate() returns TRUE if the animation series has started over at the original frame 
// for that state.  This is useful to tell when say, an explosion animation has ended, so
// the state may be changed to "dead" by the game engine.
//
// IMPORTANT! - Setting the state resets the current animation frame back to the first, since 
// different states often have different images.  For this reason, you should not set the state
// and then animate() (in that order) in the same game "tick" because you could "skip" the first
// animation frame of that "state".  Rather, animate() should be called first and then change 
// the state of the object, if necessary.
//
//---------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

//---------------------------------------------------------------------------

namespace EfromTibet {
  public struct Size {
    public int Height;
    public int Width;
  }

  public abstract class Actor {
    public Vector3 pos;             // position of object on the ski slope 
                                    // (X is how far left/right on the slope) 
                                    // (Y is how far above/below the ground)
                                    // (Z is how far down the slope)
    protected int m_state;
    protected Vector3 m_collisionSize;
    protected Vector3 m_objectSize;
    protected Size m_imgSize;       // height and width are the actual pixel height and width of the image
    
    // constructor
    public Actor() {
      // nothing to do here
    }

    public virtual Vector3 collisionSize { get { return m_collisionSize; } }
    public virtual Vector3 objectSize { get { return m_objectSize; } }
    public virtual Size imgSize { get { return m_imgSize; } }

    public abstract int state { get; set; }              // get/set definitions will be done by derived classes
    public abstract bool initialize(ContentManager c);   // initializes this object 
    public abstract bool draw(SpriteBatch spriteHandler, Rectangle? dest);  // draws the object at passed destination.  If destination is not included, it's up to the derived object to figure out what to do
    public abstract bool animate(GameTime gameTime);     // update animation frame (True, means animation has reset)
  }
}

//---------------------------------------------------------------------------
