//
// ControllerEngine.cs
//
// This class contains the XNA event handling code such as Draw() and Update().
// This class delegates work to the classes that inherit from EngineHandler.cs.
//   This way, we can separate our game and menu related code.
//
// ControllerEngine is responsible for taking care of tasks that both of the
//   EngineHandler subclasses would need to perform.  These are tasks like
//   initializing and polling the input devices, keeping track of window 
//   dimensions, etc.
//

#region Using Directives
using System;
using System.Threading;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;
#endregion

namespace EfromTibet {
  //
  // This is the main type for your game
  //
  public class ControllerEngine : Microsoft.Xna.Framework.Game {




    #region Game State Definitions/Functions
    public enum GameState {
      None,
      Menu,
      Playing
    }

    public bool switchToMenu() {
      if (m_gameState != GameState.Menu) {
        if (m_gameState == GameState.Playing) {
          m_playEngine.Deactivate();
        }
        m_menuEngine.Activate();
        m_gameState = GameState.Menu;
        return true;
      }
      return false;
    }
    public bool switchToPlaying() {
      if (m_gameState != GameState.Playing) {
        if (m_gameState == GameState.Menu) {
          m_menuEngine.Deactivate();
        }
        m_playEngine.Activate();
        m_gameState = GameState.Playing;
        return true;
      }
      return false;
    }
    public void exitGame() {
      this.Exit();
    }
    #endregion
    



    #region DEFS
    public const int CAMERA_POLL_INTERVAL = 250;
    
    protected bool m_initSuccess = false;
    public bool initSuccess { get{ return m_initSuccess; } }
    protected String m_initErrorMessage;
    public String initErrorMessage { get { return m_initErrorMessage; } }
    
    protected GraphicsDeviceManager m_graphics;
    public GraphicsDeviceManager graphics { get{ return m_graphics; } }
    
    protected Viewport m_windowSize;
    public Viewport windowSize { get{ return m_windowSize; } }

    protected KeyboardHandler m_keyboard;
    public KeyboardHandler keyboard{ get{ return m_keyboard; } }

    protected MouseHandler m_mouse;
    public MouseHandler mouse{ get{ return m_mouse; } }

    protected OpenCVProxy m_camera;
    public OpenCVProxy camera{ get{ return m_camera; } }

    protected bool m_cameraPollSuccess;
    public bool cameraPollSuccess { get { return m_cameraPollSuccess; } }

    protected SpriteBatch m_spriteHandler;
    public SpriteBatch spriteHandler{ get { return m_spriteHandler; } }

    protected GameState m_gameState;
    public GameState gameState { get { return m_gameState; } }

    protected Random m_randomGenerator;
    public Random randomGenerator { get { return m_randomGenerator; } }

    protected int timerCounter;

    protected MenuEngineHandler m_menuEngine;
    public MenuEngineHandler menuEngine { get { return m_menuEngine; } }

    protected PlayEngineHandler m_playEngine;
    public PlayEngineHandler playEngine { get { return m_playEngine; } }

    public enum InputType {
      Undefined,
      Camera,
      Mouse,
      Keyboard
    }

    protected InputType m_inputToUse;
    public InputType inputToUse { get { return m_inputToUse; } set { m_inputToUse = value; } }



    // audio stuff
    protected AudioEngine m_audioEngine;
    public AudioEngine audioEngine { get { return m_audioEngine; } }
    
    protected WaveBank m_waveBank;
    public WaveBank waveBank { get { return m_waveBank; } }

    protected SoundBank m_soundBank;
    public SoundBank soundBank { get { return m_soundBank; } }

    protected AudioCategory m_musicCategory;
    public AudioCategory musicCategory { set { m_musicCategory = value; } get { return m_musicCategory; } }


    // font to use if we need to draw error messages
    protected SpriteFont errorFont;
    #endregion




    #region Constructor
    
    // constructor
    public ControllerEngine() {
      Content.RootDirectory = "Content";
      this.IsMouseVisible = true;

      // create our graphics object
      m_graphics = new GraphicsDeviceManager(this);

      // set our preferred window dimentions
      m_graphics.PreferredBackBufferWidth = Program.NATIVE_WIDTH;
      m_graphics.PreferredBackBufferHeight = Program.NATIVE_HEIGHT;
    }

    #endregion




    // Called by the XNA framework when we should initialize our game.
    #region Initialize()
    //
    // Allows the game to perform any initialization it needs to before starting to run.
    // This is where it can query for any required services and load any non-graphic
    // related content.  Calling base.Initialize will enumerate through any components
    // and initialize them as well.
    //
    protected override void Initialize() {
      bool menuInit = false;
      bool playingInit = false;    
    
      // start out not successfully initializing
      m_initSuccess = true;
      
      // set up keyboard handler and get initial keyboard state 
      m_keyboard = new KeyboardHandler();
      m_keyboard.Update();

      // set up mouse handler and get initial mouse state 
      m_mouse = new MouseHandler();
      m_mouse.Update();

      // creates our OpenCV proxy (for camera input)
      try {
        m_camera = new OpenCVProxy();
      } catch {
        m_camera = null;
      }
      timerCounter = 0;

      // input type to use is not yet defined
      m_inputToUse = InputType.Undefined;
      
      // no initial state (yet, we will switch to one)
      m_gameState = GameState.None;

      // initialize random number generator
      m_randomGenerator = new Random();

      // fire up audio system
      try {
        m_audioEngine = new AudioEngine("Content\\Audio\\efrotAudio.xgs");
        m_waveBank = new WaveBank(audioEngine, "Content\\Audio\\Wave Bank.xwb");
        m_soundBank = new SoundBank(audioEngine, "Content\\Audio\\Sound Bank.xsb");
      } catch {
        m_initSuccess = false;
        m_initErrorMessage = "Could not initialize audio system.";
      }

      // initialize our menu and play objects
      m_menuEngine = new MenuEngineHandler();
      menuInit = m_menuEngine.Initialize();
      m_playEngine = new PlayEngineHandler();
      playingInit = m_playEngine.Initialize();

      if (m_initSuccess == true) {
        // menu initialize successfully
        if (menuInit == true) {
          if (playingInit == true) {
            m_initSuccess = true;
            m_graphics.ToggleFullScreen();

            // playingInit == false
          } else {
            m_initSuccess = false;
            m_initErrorMessage = "Could not initialize PlayEngineHandler.";
          }
          // menuInit == false
        } else {
          m_initSuccess = false;
          m_initErrorMessage = "Could not initialize MenuEngineHandler.";
        }
      }
      base.Initialize();
    }
    #endregion




    // Called by the XNA framework when we should load our content.
    #region LoadContent()
    protected override void LoadContent() {
      // Create a new SpriteBatch, which can be used to draw textures.
      m_spriteHandler = new SpriteBatch(GraphicsDevice);

      // Remember screen size
      m_windowSize = graphics.GraphicsDevice.Viewport;

      try {
        errorFont = Content.Load<SpriteFont>("ErrorText");
      } catch { }

      // if we've initialized successfully so far...
      if (m_initSuccess == true) {
        bool playLoad;
        bool menuLoad;

        menuLoad = m_menuEngine.LoadContent();
        playLoad = m_playEngine.LoadContent();

        if (menuLoad == false) {
          m_initSuccess = false;
          m_initErrorMessage += "\nCould not load menu content.";
        }
        if (playLoad == false) {
          m_initSuccess = false;
          m_initErrorMessage += "\nCould not load play content.";
        }
        if (menuLoad && playLoad) {
          m_initSuccess = true; 
          this.switchToMenu();
        }
      } // if not initialized successfully so far, we don't do anything
    }
    #endregion




    // Called by the XNA framework when we should unload our content
    #region UnloadContent
    protected override void UnloadContent() {
      m_menuEngine.UnloadContent();
      m_playEngine.UnloadContent();
    }
    #endregion




    // Called by the XNA framework when we should update.
    #region Update()
    protected override void Update(GameTime gameTime) {
      // ********************************************************
      // ****When we update, we first poll our input devices.****
      
      // poll our input devices

      Thread t = new Thread(m_camera.poll);
      if (m_camera != null) {
        if (m_camera.isInitialized()) {
          timerCounter += gameTime.ElapsedGameTime.Milliseconds;
          if (timerCounter >= CAMERA_POLL_INTERVAL) {
            timerCounter = 0;
            t.Start();
          }
        }
      }

      m_keyboard.Update();
      m_mouse.Update();
      // ********************************************************


      // allows ALT-ENTER switching to fullscreen and back
      if ((m_keyboard.IsKeyDown(Keys.LeftAlt)) || (m_keyboard.IsKeyDown(Keys.RightAlt))) {
        if (m_keyboard.WasKeyPressed(Keys.Enter)) {
          m_graphics.ToggleFullScreen();
        }
      }


      // now that that's done, we call update on the current engine object that needs it
      if (m_gameState == GameState.Playing) {
        m_playEngine.Update(gameTime);
      } else if (m_gameState == GameState.Menu) {
        m_menuEngine.Update(gameTime);

        // else no valid state, we do this
      } else {
        // nothing right now
      }


      // appease our base class
      base.Update(gameTime);
    }
    #endregion




    // Called by the XNA framework when we need to draw.
    #region Draw()
    protected override void Draw(GameTime gameTime) {
      if (m_gameState == GameState.Menu) {
        m_menuEngine.Draw(gameTime);
      } else if (m_gameState == GameState.Playing) {
        m_playEngine.Draw(gameTime);
      } else {
        // Clear background
        m_graphics.GraphicsDevice.Clear(Color.Black);

        // Start rendering sprites
        m_spriteHandler.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Deferred, SaveStateMode.SaveState);

        if (errorFont != null) {
          m_spriteHandler.DrawString(errorFont, "There was an error initializing the game:\n" + m_initErrorMessage + "\n\n\nPress ALT-F4 to exit.", new Vector2(0.0f, 0.0f), Color.White);
        }

        // We are done, draw everything on screen with help of the end method.
        m_spriteHandler.End();
      }

      base.Draw(gameTime);
    }
    #endregion




    // called when game is shutting down
    #region GameShutdown()
    public void GameShutdown() {
      if (m_camera.isInitialized()) {
        m_camera.deactivate();
      }
    }
    #endregion



/*
    #region Sprite handling

    public void RenderSprite(Texture2D texture, Rectangle rect, Rectangle? sourceRect, Color color) {

//      spriteBatch.Draw(backgroundTexture,
//        new Rectangle(0, 0, width, height),
//        Color.LightGray);

    } // RenderSprite(texture, rect, sourceRect, color)

    public void RenderSprite(Texture2D texture, Rectangle rect, Rectangle? sourceRect) {
      RenderSprite(texture, rect, sourceRect, Color.White);
    } // RenderSprite(texture, rect, sourceRect)

    public void RenderSprite(Texture2D texture, int x, int y, Rectangle? sourceRect, Color color) {
      RenderSprite(texture, new Rectangle(x, y, sourceRect.Value.Width, sourceRect.Value.Height), sourceRect, color);
    } // RenderSprite(texture, rect, sourceRect)

    public void RenderSprite(Texture2D texture, int x, int y, Rectangle? sourceRect) {
      RenderSprite(texture, new Rectangle(x, y, sourceRect.Value.Width, sourceRect.Value.Height), sourceRect, Color.White);
    } // RenderSprite(texture, rect, sourceRect)

    public void RenderSprite(Texture2D texture, Rectangle rect, Color color) {
      RenderSprite(texture, rect, null, color);
    } // RenderSprite(texture, rect, color)

    public void RenderSprite(Texture2D texture, Rectangle rect) {
      RenderSprite(texture, rect, null, Color.White);
    } // RenderSprite(texture, rect)

    public void RenderSprite(Texture2D texture) {
      RenderSprite(texture, new Rectangle(0, 0, 1024, 768), null, Color.White);
    } // RenderSprite(texture)
    #endregion
    */
  }
}
