//
// CourseSP.cs
//
// This class sets up a space-partitioning scheme for our game's StaticActor instances.
//   Using the space partitioning scheme, we can process many more objects on our course 
//   than if we had just one HUGE array of everything.
//
// To use this class:
//   1.  Create an instance of the class (default constructor).
//   2.  Call initialize() with the MAXIMUM Z distance that an object
//         can be located at.
//   3.  Call add() to add StaticActor instances to the course.  They will be 
//         placed in the proper partitions automatically.
//   4.  After adding() as many objects as desired, be sure to call the 
//         sortFarthestToNearest() method ONCE so that the objects are
//         drawn in the correct order while iterating through the objects
//         in each partition.
//   5.  Use the begin(int pID) and next() methods to iterate over the objects in 
//         the partition 'pID'.  When begin() or next() returns null, there are no
//         more objects.
//   6.  Use the getPartitionLocatedIn() method to determine which partition an object
//         or Z position is located in.  Note that higher partition numbers have FURTHER
//         Z values, so you can add to the value returned to get items in the distance.
//

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace EfromTibet {
  public class CourseSP {

    #region Constants
    protected const int DEF_PARTITION_SIZE = 5000;   // warning!  Changing this number could affect collision along the Z axis!!!
    protected const int INITIAL_PARTITION_CAPACITY = 500;
    #endregion




    #region Properties
    public int numPartitions {
      get {
        return m_numPartitions;
      }
    }

    public int partitionSize {
      get {
        return m_partitionSize;
      }
    }
    #endregion
    



    #region Protected Definitions
    protected int m_numPartitions;
    protected int m_partitionSize;
    protected StaticActor[][] m_partitionObjects;
    protected int[] m_numUsed;
    protected int m_currentIteratorPartition;
    protected int m_currentIteratorPosition;
    #endregion




    #region Constructor
    // constructor
    public CourseSP() {
      // nothing here, initialization work done in initialize()
    }
    #endregion




    #region public bool initialize(int maxPos)
    public bool initialize(int maxPos){
      m_partitionSize = DEF_PARTITION_SIZE;
      m_numPartitions = (int)Math.Ceiling((double)(maxPos) / m_partitionSize);
      m_partitionObjects = new StaticActor[m_numPartitions][];
      m_numUsed = new int[m_numPartitions];
      for (int i = 0; i < m_numUsed.Length; i++) {
        m_numUsed[i] = -1;
      }

      m_currentIteratorPartition = -1;
      m_currentIteratorPosition = -1;
      return true;
    }
    #endregion




    // ***public void sortFarthestToNearest()***
    // Sorts all of the StaticActors in each partition from 
    //   their farthest Z position to nearest Z partition.
    //
    // This allows us to draw in order, and all of the objects end up
    //   drawing so near ones cover up far ones.
    // 
    // Implementation Note:  Uses quicksort.
    //
    #region public void sortFarthestToNearest()
    public void sortFarthestToNearest() {
      for (int i = 0; i < m_numPartitions; i++) {
        sortFarthestToNearest(i);
      }
    }
    #endregion




    // ***public void sortFarthestToNearest(int partitionID)***
    // Sorts the StaticActors in 'partitionID' so that they are ordered
    //   from farthest Z position to nearest Z partition.
    //
    // This allows us to draw this partition in order, and all of the objects end up
    //   drawing so near ones cover up far ones.
    // 
    // Implementation Note:  Uses quicksort.
    //
    #region public void sortFarthestToNearest(int partitionID)
    public void sortFarthestToNearest(int partitionID) {
      if ((partitionID < 0) || (partitionID >= m_numPartitions)){
        return;
      }
      StaticActor[] arrayToSort = m_partitionObjects[partitionID];
      int numUsed = m_numUsed[partitionID];
      this.sort(arrayToSort, numUsed);
      return;
    }
    #endregion




    #region public bool add(StaticActor ca)
    public bool add(StaticActor ca) {
      int destPartitionID = this.getPartitionLocatedIn(ca);
      if (destPartitionID < 0) return false;

      StaticActor[] partitionObjectsArray = m_partitionObjects[destPartitionID];
      if (partitionObjectsArray == null) {
        partitionObjectsArray = new StaticActor[INITIAL_PARTITION_CAPACITY];
        m_partitionObjects[destPartitionID] = partitionObjectsArray;
        m_numUsed[destPartitionID] = 0;
      }

      int destLocation = m_numUsed[destPartitionID];
      if (destLocation >= partitionObjectsArray.Length) {
        int newLen = partitionObjectsArray.Length + INITIAL_PARTITION_CAPACITY;
        partitionObjectsArray = new StaticActor[newLen];
        for (int i = 0; i < m_numUsed[destPartitionID]; i++) {
          partitionObjectsArray[i] = m_partitionObjects[destPartitionID][i];
        }
        m_partitionObjects[destPartitionID] = partitionObjectsArray;
      }
      partitionObjectsArray[destLocation] = ca;
      m_numUsed[destPartitionID]++;
      return true;
    }
    #endregion




    #region public StaticActor begin(int partitionID)
    public StaticActor begin(int partitionID) {      
      if ((partitionID < m_numPartitions) && (partitionID >= 0)) {
        if (m_numUsed[partitionID] <= 0) {
          return null;
        } else {
          m_currentIteratorPartition = partitionID;
          m_currentIteratorPosition = 0;
          return m_partitionObjects[partitionID][0];
        }
      } else {
        return null;
      }
    }
    #endregion




    #region public StaticActor next()
    public StaticActor next() {
      if (m_currentIteratorPartition < 0) {
        return null;
      }
      m_currentIteratorPosition++;
      if (m_currentIteratorPosition < m_numUsed[m_currentIteratorPartition]) {
        return m_partitionObjects[m_currentIteratorPartition][m_currentIteratorPosition];
      } else {
        m_currentIteratorPosition = -1;
        m_currentIteratorPartition = -1;
        return null;
      }
    }
    #endregion




    #region public int getPartitionLocatedIn overrides
    public int getPartitionLocatedIn(int pos) {
      //int location = (int)(Math.Ceiling((double)pos / m_partitionSize));
      int location = pos / m_partitionSize;
      if (location >= m_numPartitions) {
        return -1;
      } else {
        return location;
      }
    }

    public int getPartitionLocatedIn(Vector3 pos) {
      return this.getPartitionLocatedIn((int)(pos.Z));
    }

    public int getPartitionLocatedIn(Actor a) {
      return this.getPartitionLocatedIn((int)(a.pos.Z));
    }
    #endregion




    #region Quicksort Functions
    //---------------------------------------------------------------------------
    // performs the quick sort
    protected void sort(StaticActor[] arr, int length){
      Quick_Sort_Main(arr, 0, length-1);
      return;
    }

    //---------------------------------------------------------------------------
    // private helper function for the quicksort routine
    protected void Quick_Sort_Main(StaticActor[] values, int first, int last){
      if (first < last){
        int splitPoint = 0;

        split(values, first, last, ref splitPoint);
        Quick_Sort_Main(values, first, splitPoint-1);
        Quick_Sort_Main(values, splitPoint+1, last);
      }
    }

    //---------------------------------------------------------------------------
    // private helper function for the quicksort routine
    protected void split(StaticActor[] values, int first, int last, ref int splitPoint){
      StaticActor temp;
      StaticActor splitVal = values[first];
      int saveFirst = first;
      bool onCorrectSide;

      first++;
      do{
        onCorrectSide = true;
        while (onCorrectSide){          // Move first toward last
          //if (values[first] > splitVal){
          if (isHigherThan(values[first], splitVal) == true){
            onCorrectSide = false;
          }else{
            first++;
            onCorrectSide = (first <= last);
          }
        }

        onCorrectSide = (first <= last);
        while (onCorrectSide){           // Move last toward first

          //if (values[last] <= splitVal){
          if (isHigherThan(values[last], splitVal) == false){
            onCorrectSide = false;
          }else{
            last--;
            onCorrectSide = (first <= last);
          }
        }


        if (first < last){
          //Swap(values[first], values[last]);
          temp = values[first];
          values[first] = values[last];
          values[last] = temp;

          first++;
          last--;
        }
      } while (first <= last);

      splitPoint = last;

      //Swap(values[saveFirst], values[splitPoint]);
      temp = values[saveFirst];
      values[saveFirst] = values[splitPoint];
      values[splitPoint] = temp;
    }

    //---------------------------------------------------------------------------
    // returns true if item1 should be ABOVE item2 in the sorted list
    //
    // --> true if item1 above item2
    // --> false if item1 below/equal item2
    protected bool isHigherThan(StaticActor item1, StaticActor item2){
      if (item1.pos.Z < item2.pos.Z) {
        return true;
      } else {
        return false;
      }
    }
    #endregion




  }//end CourseSP
}