using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace EfromTibet {
  public class PlayEngineHandler : EngineHandler {
    
    public enum PlayState {
      Beginning,      // about to switch to Playing state (running intro to level sequence)
      Playing,        // playing game
      Paused,         // game is paused
      GameOver,       // gameover
      Complete        // course is complete
    }




    #region Constants
    protected const int BEGINNINGINTROTIME = 7000;

    // how far we can "see".  If objects are further than this
    // distance, the engine does not attempt to draw them.  This way, we avoid 
    // drawing any "specs" where objects are interpolated down to one pixel.
    protected const int MAX_HORIZON_VIEW_DISTANCE = 35000;

    // the percentage of the screen that the snow landscape occupies.
    // We use this percentage to calculate where the horizon line is located.
    protected const double HORIZON_PIXELS_PERCENTAGE = 0.252604167;

    // the width of the "course"
    protected const int COURSE_WIDTH = 40000;

    // the "length" of the course
    public const int COURSE_MAX_Z = 1000000;

    // minimum Z position
    public const int COURSE_MIN_Z = -10000;

    protected const int SNOWMAN_X_DIST_APART = COURSE_WIDTH / 5;

    // number of static course objects to generate
    public const int DEF_NUM_STATIC_ACTORS_TO_GENERATE = 4500;// default number of static actors to generate for the course
    public const int STATIC_ACTORS_TO_GENERATE_STEP = 500;    // step for when we're adjusting the number to generate

    // player stats
    protected const int PLAYER_INITIAL_SPEED = 2000;          // the initial speed of the player
    protected const int PLAYER_MAX_SPEED = 8000;              // maximum speed of the player
    protected const int PLAYER_MIN_SPEED = -500;              // mimimum speed of the player
    protected const int PLAYER_MAX_ACCELERATION_PER_SEC = 20; // maximum acceleration of the player per second
    protected const int PLAYER_MAX_DECELERATION_PER_SEC = 45; // maximum deceleration of the player per second
    protected const int PLAYER_MAX_ANGLE = 35;                // maximum angle player can turn to the right
    protected const int PLAYER_MIN_ANGLE = -35;               // maximum angle the player can turn to the left
    protected const float PLAYER_DEGREES_PER_SEC = 90f;       // how many degress the player can turn in a single second (we do this to have a delay when changing directions)
    protected const int PLAYER_JUMP_SPEED = 1750;

    protected const int GRAVITY_ACCELERATION_PER_SEC = 3500;

    protected const float PLAYER_PAIN_FACTOR_MULTIPLIER = 0.0040f;

    protected const int CAMERA_DECELERATION_RATE = 125;
    #endregion




    #region Variable Defs
    // state of our game (paused, playing, game over, etc)
    protected PlayState m_playState;

    // keep track of time for our game
    protected TimeSpan m_stateTimer;
    protected TimeSpan m_playTimer;
    
    // background texture (sky)
    protected Texture2D m_skyBackground;
    protected Rectangle m_skyBackground_dest;

    // snow texture
    protected Texture2D m_snow;

    // border for face locator
    protected Texture2D m_border;
    protected Rectangle m_border_dest;
    // face overlay to draw inside border
    protected Texture2D m_faceOverlay;

    // our font objects
    protected SpriteFont m_CountdownFont;
    protected SpriteFont m_mediumFont;


    // our player object instance
    protected Player m_player;

    // course objects that don't move
    protected CourseSP m_staticCourseObjects;

    // y pixel position of the horizon line
    protected int horizonYPixelPos;

    // position of our "camera" (we don't always follow the player)
    protected Vector3 m_cameraPosition;
    // speed of our camera
    protected float m_cameraSpeed;

    protected Cue songPlaying;

    // the number of static actors to generate for the course
    public int numStaticActorsToGenerate = DEF_NUM_STATIC_ACTORS_TO_GENERATE;

    #endregion




    #region Constructor
    public PlayEngineHandler() {

    }
    #endregion




    #region Initialize()
    public override bool Initialize() {
      return true;
    }
    #endregion




    #region LoadContent()
    public override bool LoadContent() {
      try {
        ContentManager c = Program.controller.Content;

        m_skyBackground = c.Load<Texture2D>("Sky");
        m_skyBackground_dest = new Rectangle(0, 0, Program.controller.windowSize.Width, Program.controller.windowSize.Height);

        m_border = c.Load<Texture2D>("border");
        m_border_dest = new Rectangle((Program.controller.windowSize.Width) - (int)(285 * (float)(Program.controller.windowSize.Width) / Program.NATIVE_WIDTH), (Program.controller.windowSize.Height) - (int)(770 * (float)(Program.controller.windowSize.Height) / Program.NATIVE_HEIGHT), (int)(280 * (float)(Program.controller.windowSize.Width) / Program.NATIVE_WIDTH),(int)(150 * (float)(Program.controller.windowSize.Height) / Program.NATIVE_HEIGHT));

        m_snow = c.Load<Texture2D>("Snow");

        m_CountdownFont = c.Load<SpriteFont>("CountdownText");
        m_mediumFont = c.Load<SpriteFont>("PlayMediumText");


        m_faceOverlay = c.Load<Texture2D>("smile");


        bool playerInitSuccess;
        m_player = new Player();
        playerInitSuccess = m_player.initialize(c);
        if (playerInitSuccess == false){
          return false;
        }


        // generate the course
        if (generateCourseActors(c) == false) {
          return false;
        }


        // calculate horizon pixel Y position (in real screen pixels)
        horizonYPixelPos = (int)((Program.controller.windowSize.Height * 3.0 / 4));

        return true;
      } catch {
        return false;
      }
    }
    #endregion




    #region UnloadContent()
    public override bool UnloadContent() {
      return true;
    }
    #endregion




    #region Update()
    public override void Update(GameTime gameTime) {
      ControllerEngine controller = Program.controller;




      // *********************************************************************************
      // ***PlayState.Beginning***
      // *
      // * All we have to do here is count up our state timer, so the 3, 2, 1, GO! message works.
      // *
      if (m_playState == PlayState.Beginning) {
        m_stateTimer = m_stateTimer.Add(gameTime.ElapsedGameTime);
        if (m_stateTimer.TotalMilliseconds >= BEGINNINGINTROTIME) {
          m_playState = PlayState.Playing;
          m_stateTimer = m_stateTimer.Add(m_stateTimer.Negate());
          controller.musicCategory = controller.audioEngine.GetCategory("Music");
          songPlaying = controller.soundBank.GetCue("JamesBondTechno");
          controller.musicCategory.SetVolume(0.5f);
          songPlaying.Play();
        }
      }
      // ***End of PlayState.Beginning***
      // *********************************************************************************




      // *********************************************************************************
      // ***PlayState.Playing***
      // *
      // * In this state, we have to do a lot of actions to make the game work.
      // *  1.  Count up our game timer.
      // *  2.  If P was pressed, go to PAUSED state (takes effect next time around.)
      // *  3.  If the player object is currently alive:
      // *     3.1.  Process camera user input, if we're using camera input.
      // *     3.2.  Process mouse user input, if we're using mouse input.
      // *     3.3.  Process keyboard input, if we're using keyboard input.
      // *     3.4.  Based on that input, adjust speed and angle of player.
      // *  4.  Move the player's position based on speed, angle, and vSpeed of player.
      // *  5.  Update the camera position.
      // *  6.  Animate the player.
      // *  7.  Do collision detection between the player and static objects on the course.
      // *
      else if (m_playState == PlayState.Playing) {

        // count up our game timer
        m_playTimer = m_playTimer.Add(gameTime.ElapsedGameTime);

        // do we need to pause the game
        if (controller.keyboard.WasKeyPressed(Keys.P)) {
          m_playState = PlayState.Paused;
        }

        // if player object can be controlled (in the moving state)
        //   player object might not be able to be controlled, say, when he's tripping, etc
        if (m_player.state == Player.STATE_MOVING) {

          float accelerationFactor = 0; // percentage of maximum acceleration (or deceleration) the user wants to increase/decrease speed by
          float anglePercentage = 0;    // percentage of max (or min) angle the user wants to face
          bool doJump = false;

          if (controller.inputToUse == ControllerEngine.InputType.Mouse) {
            // mouse input
            //   
            //   angle depends on where the current X position of the mouse is.
            //   mouse center = angle of zero
            //   mouse all the way left = MIN angle
            //   mouse all the way right = MAX angle
            anglePercentage = 2 * ((float)(controller.mouse.X) / controller.windowSize.Width - 0.5f);
            if (anglePercentage > 1) anglePercentage = 1.0f;
            if (anglePercentage < -1) anglePercentage = -1.0f;
            //
            //   acceleration depends on where the current Y position of the mouse is.
            //   mouse center = no acceleration
            //   mouse all the way top = MAX acceleration
            //   mouse all the way bottom = MAX deceleration
            accelerationFactor = -2 * ((float)(controller.mouse.Y) / controller.windowSize.Height - 0.5f);
            if (accelerationFactor > 1) accelerationFactor = 1.0f;
            if (accelerationFactor < -1) accelerationFactor = -1.0f;

            // do we need to jump?
            if (controller.mouse.WasLeftButtonPressed()) {
              doJump = true;  // set flag
            }
          } else

          if (controller.inputToUse == ControllerEngine.InputType.Keyboard) {
              
            // keyboard input
            //   keyboard input is limited to either 1 or 0 for percentages.
            //   Thus, the target angle for keyboard input is always 0, the 
            //     maximum angle, or the minimum angle.
            //   Likewise, the acceleration for keyboard input is always 0,
            //     the maximum acceleration, or the maximum deceleration.
            //
            //   Mouse or camera input allows more fine-tuned control
            if ((controller.keyboard.IsKeyDown(Keys.Right)) || (controller.keyboard.IsKeyDown(Keys.D))) {
              anglePercentage = 1.0f;
            }
            if ((controller.keyboard.IsKeyDown(Keys.Left)) || (controller.keyboard.IsKeyDown(Keys.A))){
              anglePercentage = -1.0f;
            }
            if ((controller.keyboard.IsKeyDown(Keys.Up)) || (controller.keyboard.IsKeyDown(Keys.W))){
              accelerationFactor = 1.0f;
            }
            if ((controller.keyboard.IsKeyDown(Keys.Down)) || (controller.keyboard.IsKeyDown(Keys.S))) {
              accelerationFactor = -1.0f;
            }
            // jump?
            if (controller.keyboard.WasKeyPressed(Keys.Space)) {
              doJump = true;  // set flag
            }
          }else 

          if (controller.inputToUse == ControllerEngine.InputType.Camera) {
            FaceRecord fi = controller.camera.faceInfo;
            // mouse input
            //   
            //   angle depends on where the current X position of the mouse is.
            //   mouse center = angle of zero
            //   mouse all the way left = MIN angle
            //   mouse all the way right = MAX angle
            if (FaceRecord.currentlyRecognized == true && controller.inputToUse == ControllerEngine.InputType.Camera)
            {
                anglePercentage = ((float)(FaceRecord.x + FaceRecord.width / 2)) / FaceRecord.camWidth;
                anglePercentage -= 0.5f;
                anglePercentage *= 4;

                anglePercentage = -anglePercentage;

                if (anglePercentage > 1) anglePercentage = 1.0f;
                if (anglePercentage < -1) anglePercentage = -1.0f;
                //
                //   acceleration depends on where the current Y position of the mouse is.
                //   mouse center = no acceleration
                //   mouse all the way top = MAX acceleration
                //   mouse all the way bottom = MAX deceleration
                accelerationFactor = ((float)(FaceRecord.width) / FaceRecord.camWidth);
                accelerationFactor -= 0.25f;
                accelerationFactor *= 2;
                if (accelerationFactor > 1) accelerationFactor = 1.0f;
                if (accelerationFactor < -1) accelerationFactor = -1.0f;

                // do we need to jump? 
                if (FaceRecord.clicked)
                {
                    doJump = true;  // set flag
                }
            }
          } 

          // if player is on the ground
          if (m_player.vSpeed == 0) {
            // calculate target angle for player
            float targetAngle = 0;
            if (anglePercentage > 0) {
              targetAngle = anglePercentage * PLAYER_MAX_ANGLE;
            } else if (anglePercentage < 0) {
              targetAngle = anglePercentage * PLAYER_MIN_ANGLE * -1;
            }

            if (m_player.speed < 0) targetAngle = -targetAngle;

            // if we need to turn more to the RIGHT,
            if (targetAngle > m_player.angle) {
              m_player.angle += (float)(PLAYER_DEGREES_PER_SEC * gameTime.ElapsedGameTime.TotalMilliseconds / 1000);

              // don't overshoot
              if (targetAngle < m_player.angle) {
                m_player.angle = targetAngle;
              }
              // turn more to the left
            } else if (targetAngle < m_player.angle) {
              m_player.angle -= (float)(PLAYER_DEGREES_PER_SEC * gameTime.ElapsedGameTime.TotalMilliseconds / 1000);

              // don't overshoot
              if (targetAngle > m_player.angle) {
                m_player.angle = targetAngle;
              }

            }


            // increase/decrease our speed if necessary
            if (accelerationFactor > 0) {
              m_player.speed += accelerationFactor * PLAYER_MAX_ACCELERATION_PER_SEC;
              // don't go past maximum
              if (m_player.speed > PLAYER_MAX_SPEED) {
                m_player.speed = PLAYER_MAX_SPEED;
              }
            } else if (accelerationFactor < 0) {
              m_player.speed += accelerationFactor * PLAYER_MAX_DECELERATION_PER_SEC;
              // don't go slower than minimum
              if (m_player.speed < PLAYER_MIN_SPEED) {
                m_player.speed = PLAYER_MIN_SPEED;
              }
            }


            // if we need to make the player jump
            if (doJump == true) {
              m_player.vSpeed = PLAYER_JUMP_SPEED;
            }

          }
        } // endif (player.state == Player.STATE_ALIVE)

        //***********************************************************************
        // At this point we've processed input.
        //***********************************************************************




        // from our speed and angle numbers, calculate velocity vector
        Vector3 v;
        v.X = (float)(m_player.speed * Math.Sin(m_player.angle * Math.PI / 180));
        v.Z = (float)(m_player.speed * Math.Cos(m_player.angle * Math.PI / 180));

        m_player.pos.Z += (int)(v.Z * gameTime.ElapsedGameTime.TotalMilliseconds / 1000);
        m_player.pos.X += (int)(v.X * gameTime.ElapsedGameTime.TotalMilliseconds / 1000);

        // wrap around the course
        if (m_player.pos.X > COURSE_WIDTH) {
          m_player.pos.X -= COURSE_WIDTH;
        }

        // wrap around the course
        if (m_player.pos.X < 0) {
          m_player.pos.X += COURSE_WIDTH;
        }



        // move us down if we are in the air
        if (m_player.pos.Y > 0) {
          m_player.vSpeed -= (float)((float)gameTime.ElapsedGameTime.TotalMilliseconds / 1000 * GRAVITY_ACCELERATION_PER_SEC);
        }
        m_player.pos.Y += (float)(gameTime.ElapsedGameTime.TotalMilliseconds / 1000 * m_player.vSpeed);
        if (m_player.pos.Y < 0) {
          m_player.pos.Y = 0;
          m_player.vSpeed = 0;
        }

        // prevent player from going backwards right at the top
        if (m_player.pos.Z < COURSE_MIN_Z) {
          m_player.pos.Z = COURSE_MIN_Z;
        }

        // if player has completed the course
        if (m_player.pos.Z >= COURSE_MAX_Z) {
          // player is still alive
          if (m_player.state == Player.STATE_MOVING) {
            m_playState = PlayState.Complete;
          }
        }

        //***********************************************************************
        // At this point we've moved the player.
        //***********************************************************************




        // how we update the camera position depends on the player state
        if (m_player.state == Player.STATE_MOVING) {
          moveCameraWithPlayer();
        } else {
          m_cameraPosition.Z += (float)(m_cameraSpeed * gameTime.ElapsedGameTime.TotalMilliseconds / 1000);
          m_cameraSpeed -= CAMERA_DECELERATION_RATE;
          if (m_cameraSpeed < 0) {
            m_cameraSpeed = 0;
          }
          m_player.angle = 0;
          m_player.speed -= CAMERA_DECELERATION_RATE / 2;
          if (m_player.speed < 0) {
            m_player.speed = 0;
          }
        }


        bool justGotUp = false;

        // do player animation stuff
        switch (m_player.state) {
          case Player.STATE_MOVING:
            m_player.animate(gameTime);
            break;
          case Player.STATE_FALLING:
            if (m_player.animate(gameTime) == true) {
              m_player.state = Player.STATE_GETTINGUP;
            }
            break;
          case Player.STATE_GETTINGUP:
            if (m_player.animate(gameTime) == true) {
              m_player.state = Player.STATE_MOVING;
              justGotUp = true;
            }
            break;
          case Player.STATE_DYING:
            if (m_player.animate(gameTime) == true) {
              m_player.state = Player.STATE_DEAD;
            }
            break;
          case Player.STATE_DEAD:
            m_player.animate(gameTime);
            m_playState = PlayState.GameOver;
            break;
        }






        Vector3 objectMiddle;
        BoundingBox objectBB;

        // get logical middle of the player 3D bounding box
        Vector3 playerMiddle = m_player.pos;
        playerMiddle += m_player.objectSize / 2;
        // wrap X position around, if necessary
        if (playerMiddle.X >= COURSE_WIDTH) {
          playerMiddle.X -= COURSE_WIDTH;
        }
        // from that middle point, calculate collision bounding box Min and Max points
        BoundingBox playerBB;
        playerBB.Min = playerMiddle - m_player.collisionSize / 2;
        playerBB.Max = playerMiddle + m_player.collisionSize / 2;

        // get partition numbers that both the near Z and far Z plane of our collision bounding box exist in
        // most of the time, these should be in the same partition, but we might be on a boundary
        int playerPartitionLow = m_staticCourseObjects.getPartitionLocatedIn((int)playerBB.Min.Z);
        int playerPartitionHigh = m_staticCourseObjects.getPartitionLocatedIn((int)playerBB.Max.Z);

        // check one partition behind ours
        playerPartitionLow--;

        // to one partition in front of ours
        playerPartitionHigh++;

        // loop for all partitions to check
        for (int curPartition = playerPartitionLow; curPartition <= playerPartitionHigh; curPartition++) {
          for (StaticActor thisSA = m_staticCourseObjects.begin(curPartition); thisSA != null; thisSA = m_staticCourseObjects.next()) {
            if (thisSA.canCollide == true) {
              // calculate middle of this object
              objectMiddle = thisSA.pos;
              objectMiddle += thisSA.objectSize / 2;

              // wrap X position around, if necessary
              if (objectMiddle.X >= COURSE_WIDTH) {
                objectMiddle.X -= COURSE_WIDTH;
              }



              int offset;
              offset = (int)(objectMiddle.X - playerMiddle.X);
              // find the "wrapping offset" (what our offset would be if we accounted for course wrapping)
              int woffset;
              if (offset > 0) {  // calculate wrapping in the correct direction
                woffset = offset - COURSE_WIDTH;
              } else {
                woffset = offset + COURSE_WIDTH;
              }
              // now, if our wrapping offset is less absolute distance than our normal offset,
              if (Math.Abs(woffset) < Math.Abs(offset)) {
                // move our object's middle "outside" the course X range of 0 to COURSEWIDTH
                // we do this so our bounding box collision can work if the player is RIGHT on the boundary
                // essentially, we're cheating
                if (woffset < 0) {
                  objectMiddle.X -= COURSE_WIDTH;
                } else {
                  objectMiddle.X += COURSE_WIDTH;
                }
              }


              // set up this object's bounding box
              objectBB.Min = objectMiddle - thisSA.collisionSize / 2;
              objectBB.Max = objectMiddle + thisSA.collisionSize / 2;

              // if we are colliding with this object in any way
              if (playerBB.Intersects(objectBB)) {

                // can we "ride over" this object (such as a bump, etc)
                if (thisSA.rideOver == true) {
                  // ride "over" the object

                  // percentage of the object in X and Z directions
                  float percentageOverZ = (playerMiddle.Z - objectBB.Min.Z) / (thisSA.collisionSize.Z);
                  float percentageOverX = (playerMiddle.X - objectBB.Min.X) / (thisSA.collisionSize.X);

                  // if both percentages seem OK,
                  if ((percentageOverZ >= 0) && (percentageOverZ <= 1.0) && (percentageOverX >= 0) && (percentageOverX <= 1.0)) {


                    // position the player higher, depending on where we are with respect to the center of the object
                    // towards the middle, we're at the object's max Y, but toward the edges, we reach 0 above the ground
                    float percentofY = (float)(Math.Sin(Math.PI * percentageOverZ));
                    float targetY = (percentofY * objectBB.Max.Y);
                    targetY *= (float)(Math.Sin(Math.PI * percentageOverX));

                    // only move the player if we're too low (don't move player down)
                    if (m_player.pos.Y < targetY) {
                      m_player.pos.Y = targetY;
                      m_player.vSpeed = 0;
                    }


                  }

                  // if this is an object we can "trip" over
                } else {
                  if (m_player.state == Player.STATE_MOVING) {
                    if (justGotUp == false) {
                      controller.soundBank.PlayCue("treecrash2");


                      int painFactor = (int)(m_player.speed * PLAYER_PAIN_FACTOR_MULTIPLIER);
                      if (painFactor > 0) {
                        m_player.health -= painFactor;
                      }

                      // fatal collision
                      if (m_player.health < 0) {
                        m_player.health = 0;
                        m_player.state = Player.STATE_DYING;

                      // nonfatal collision
                      } else {
                        m_player.state = Player.STATE_FALLING;
                      }

                    // if we just got up, move us in front of this object
                    } else {
                      float zdest = objectBB.Max.Z + m_player.collisionSize.Z / 2;
                      if (m_player.pos.Z < zdest) {
                        m_player.pos.Z = zdest;
                      }
                    }
                  }

                } // endif this is an object we can trip over
              }
            } // end if (thisSA.canCollide == true)
          } // for each thisSA in curPartition
        } // for each partition we need to check



      }
      // ***End of PlayState.Playing***
      // *********************************************************************************




      // *********************************************************************************
      // ***PlayState.Paused***
      // *
      // * In this state, all we do is wait till the user unpauses the game.
      // *
      else if (m_playState == PlayState.Paused) {
        if (controller.keyboard.WasKeyPressed(Keys.P) || controller.mouse.WasLeftButtonPressed()) {
          m_playState = PlayState.Playing;
        }
      }
      // ***End of PlayState.Paused***
      // *********************************************************************************




      // *********************************************************************************
      // ***PlayState.GameOver or PlayState.Complete***
      // *
      // * In this state, we don't need to do anything (currently).  Nothing is in motion by now.
      // *   We do check for a click, which if detected, takes us to the main menu.
      // *
      else if ((m_playState == PlayState.GameOver) || (m_playState == PlayState.Complete)) {
        // switch to main menu if click detected
        if (controller.mouse.WasLeftButtonPressed()) {
          controller.switchToMenu();
        }
        if ((controller.keyboard.WasKeyPressed(Keys.Enter)) || (controller.keyboard.WasKeyPressed(Keys.Space))) {
          controller.switchToMenu();
        }
      }
      // ***End of PlayState.GameOver***
      // *********************************************************************************




      // in all states ESC returns to main menu
      if (controller.keyboard.WasKeyPressed(Keys.Escape)) {
        controller.switchToMenu();
      }



      return;
    }
    
    
    
    
    protected void moveCameraWithPlayer() {
      m_cameraPosition = m_player.pos;
      m_cameraPosition.X += m_player.objectSize.X / 2;
      m_cameraPosition.Y += m_player.objectSize.Y;
      if (m_cameraPosition.X >= COURSE_WIDTH) m_cameraPosition.X -= COURSE_WIDTH;
      m_cameraSpeed = m_player.speed;
    }
    
    #endregion




    #region Draw()
    public override void Draw(GameTime gameTime) {
      // get references into local variables to save us some typing
      GraphicsDeviceManager g = Program.controller.graphics;
      SpriteBatch spriteHandler = Program.controller.spriteHandler;

      // Clear background
      g.GraphicsDevice.Clear(Color.Magenta);

      // Start rendering sprites
      spriteHandler.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Deferred, SaveStateMode.SaveState);

      


      // draw the sky
      spriteHandler.Draw(m_skyBackground, m_skyBackground_dest, Color.White);


      

      // draw snow
      Rectangle snowDest;
      snowDest.X = 0;
      snowDest.Width = Program.controller.windowSize.Width;
      snowDest.Y = horizonYPixelPos;
      snowDest.Height = Program.controller.windowSize.Height - horizonYPixelPos;
      spriteHandler.Draw(m_snow, snowDest, Color.White);

      
          

      // WE DRAW DIFFERENTLY DEPENDING ON IF OUR PLAYER IS MOVING (CAMERA IS FIXED BEHIND PLAYER)
      //   OR IF PLAYER IS FALLING, DYING, DEAD, GETTING UP, (WHERE CAMERA IS NOT FIXED BEHIND PLAYER)
      //
      //   We have to do this because if the camera is behind the player, the player will always be "in front" of
      //     all of the course objects, and should be drawn last.
      //   If the player is actually being drawn on the course (and scaled, along with all of that),
      //     we have to draw objects that are closer to the camera after we draw him.
      //      
      if (m_player.state == Player.STATE_MOVING) {
        // locate the partitions that we need to draw objects from
        int startPID = m_staticCourseObjects.getPartitionLocatedIn(m_cameraPosition);
        int stopPID = MAX_HORIZON_VIEW_DISTANCE / m_staticCourseObjects.partitionSize + startPID;
        // for each partition
        for (int curPID = stopPID; curPID >= startPID; curPID--){
          // draw all static objects for this partition
          for (StaticActor curSActor = m_staticCourseObjects.begin(curPID); curSActor != null; curSActor = m_staticCourseObjects.next() ) {
            // draw this object
            this.drawActorOnCourse(spriteHandler, curSActor, m_cameraPosition);
          } // for each object in the partition
        } // for each partition

        // calculate where we are going to draw the player image
        Rectangle playerDest;
        playerDest.Width = (int)(m_player.imgSize.Width * (float)(Program.controller.windowSize.Width) / Program.NATIVE_WIDTH);
        playerDest.Height = (int)(m_player.imgSize.Height * (float)(Program.controller.windowSize.Height) / Program.NATIVE_HEIGHT);
        playerDest.X = Program.controller.windowSize.Width / 2 - playerDest.Width / 2;
        playerDest.Y = Program.controller.windowSize.Height - (int)(playerDest.Height);
        // now, draw the player 
        m_player.draw(spriteHandler, playerDest);
      



      // if player dead, dying, falling, or getting up,
      //   (this is the condition where we have to draw the player in between drawing the course objects)
      } else {

        // locate the partitions that we need to draw objects from
        int startPID = m_staticCourseObjects.getPartitionLocatedIn(m_cameraPosition);
        int stopPID = MAX_HORIZON_VIEW_DISTANCE / m_staticCourseObjects.partitionSize + startPID;

        bool playerDrawn = false;
        int playerPID = m_staticCourseObjects.getPartitionLocatedIn(m_player);

        // for each partition
        for (int curPID = stopPID; curPID >= startPID; curPID--) {
          if (playerPID == curPID) {
            // draw all static objects for this partition
            for (StaticActor curSActor = m_staticCourseObjects.begin(curPID); curSActor != null; curSActor = m_staticCourseObjects.next()) {
              if (playerDrawn == false) {
                if (m_player.pos.Z > curSActor.pos.Z) {
                  this.drawActorOnCourse(spriteHandler, m_player, m_cameraPosition);
                  playerDrawn = true;
                }
              }
              // draw this object
              this.drawActorOnCourse(spriteHandler, curSActor, m_cameraPosition);
            } // for each object in the partition

            if (playerDrawn == false) {
              this.drawActorOnCourse(spriteHandler, m_player, m_cameraPosition);
              playerDrawn = true;
            }
          } else {
            // draw all static objects for this partition
            for (StaticActor curSActor = m_staticCourseObjects.begin(curPID); curSActor != null; curSActor = m_staticCourseObjects.next()) {
              // draw this object
              this.drawActorOnCourse(spriteHandler, curSActor, m_cameraPosition);
            } // for each object in the partition
          }
        } // for each partition    


      } // endif player dead, dying, falling, or getting up,



      // draw the camera HUD stuff, if necessary
      if (Program.controller.inputToUse == ControllerEngine.InputType.Camera) {
        // draw the face border
        spriteHandler.Draw(m_border, m_border_dest, Color.White);

        // prevent DIV0
        if ((FaceRecord.camWidth != 0) && (FaceRecord.camHeight != 0)){
          Rectangle facedisp;
          facedisp.Width = (m_border_dest.Width * FaceRecord.width / FaceRecord.camWidth);
          facedisp.Height = (m_border_dest.Height * FaceRecord.height / FaceRecord.camHeight);
          int faceMiddleX;
          int faceMiddleY;
          faceMiddleX  = FaceRecord.x + FaceRecord.width / 2;
          faceMiddleY = FaceRecord.y + FaceRecord.height / 2;
          
          // flip about vertical axis, because for some reason face coordinates are flipped
          faceMiddleX = FaceRecord.camWidth - faceMiddleX;
          
          facedisp.X = m_border_dest.Left + (m_border_dest.Width * faceMiddleX / FaceRecord.camWidth) - facedisp.Width / 2;
          facedisp.Y = m_border_dest.Top + (m_border_dest.Height * faceMiddleY / FaceRecord.camHeight) - facedisp.Height / 2;

          spriteHandler.Draw(m_faceOverlay,  facedisp, null, Color.White);
        }
      }



      const int XBUFFER_SIZE = 3;
      const int YBUFFER_SIZE = 0;

      int healthLineY;

      // draw the "timer" text
      {
        string minutesStr = ((int)(m_playTimer.TotalMinutes)).ToString();
        if (minutesStr.Length < 2) {
          minutesStr = "0" + minutesStr;
        }
        string secondsStr = m_playTimer.Seconds.ToString();
        if (secondsStr.Length < 2) {
          secondsStr = "0" + secondsStr;
        }
        String toDisplay = "Time: " + minutesStr + ":" + secondsStr;
        Vector2 timerSize = m_mediumFont.MeasureString(toDisplay);

        Vector2 textDest;
        textDest.X = XBUFFER_SIZE;//Program.controller.windowSize.Width - XBUFFER_SIZE * 2 - timerSize.X;
        textDest.Y = YBUFFER_SIZE * 2 + timerSize.Y;
        spriteHandler.DrawString(m_mediumFont, toDisplay, textDest, Color.Red);

        healthLineY = (int)(textDest.Y + timerSize.Y*2 + YBUFFER_SIZE);
      }




      // draw "press ESC to return to menu"
      {
        Vector2 textDest;
        textDest = m_mediumFont.MeasureString("Press ESC to return to the main menu.");
        textDest.X = XBUFFER_SIZE;
        textDest.Y = Program.controller.windowSize.Height - textDest.Y - YBUFFER_SIZE ;
        spriteHandler.DrawString(m_mediumFont, "Press ESC to return to the main menu.", textDest, Color.Red);
      }




      // draw the "percentage of course complete" text
      {
        int percentComplete = 100 * (int)m_player.pos.Z / COURSE_MAX_Z;
        if (percentComplete > 100) {
          percentComplete = 100;
        }else if (percentComplete < 0) {
          percentComplete = 0;
        }
        String toDisplay = "Percent of Slope Traveled: " + percentComplete.ToString() + "%";
        Vector2 size = m_mediumFont.MeasureString(toDisplay);

        Vector2 textDest;
        textDest.X = XBUFFER_SIZE;//Program.controller.windowSize.Width - XBUFFER_SIZE * 2 - size.X;
        textDest.Y = YBUFFER_SIZE;
        spriteHandler.DrawString(m_mediumFont, toDisplay, textDest, Color.Red);
      }




      // draw the "health" text
      {
        String toDisplay = "Health: " + m_player.health.ToString();
        //Vector2 size = m_mediumFont.MeasureString(toDisplay);

        Vector2 textDest;
        textDest.X = XBUFFER_SIZE;//Program.controller.windowSize.Width - XBUFFER_SIZE * 2 - size.X;
        textDest.Y = healthLineY;
        spriteHandler.DrawString(m_mediumFont, toDisplay, textDest, Color.Red);
      }




      // if BEGINNING sequence playing
      if (m_playState == PlayState.Beginning) {
        // all of this junk figures out our
        // Ready?   3, 2, 1, GO! sequence
        int countdownNumber = (int)((BEGINNINGINTROTIME - m_stateTimer.TotalMilliseconds)) / 1000;
        String numberText;
        if (countdownNumber < 1) {
          numberText = "GO!";
        } else if (countdownNumber < 4) {
          numberText = countdownNumber.ToString() + "!";
        } else {
          numberText = "Ready?";
        }
        Vector2 countdownSize = m_CountdownFont.MeasureString(numberText);
        Vector2 vecDest;
        vecDest.X = Program.controller.windowSize.Width / 2 - countdownSize.X / 2;
        vecDest.Y = Program.controller.windowSize.Height / 2 - countdownSize.Y / 2;
        spriteHandler.DrawString(m_CountdownFont, numberText, vecDest, Color.Red);
      
      // if GAMEOVER
      }else if (m_playState == PlayState.GameOver) {
        Vector2 gameoverSize = m_CountdownFont.MeasureString("GAME OVER");
        Vector2 vecDest;
        vecDest.X = Program.controller.windowSize.Width / 2 - gameoverSize.X / 2;
        vecDest.Y = Program.controller.windowSize.Height / 2 - gameoverSize.Y / 2;
        spriteHandler.DrawString(m_CountdownFont, "GAME OVER", vecDest, Color.Red);
      }

      // if WON THE GAME
      else if (m_playState == PlayState.Complete) {
        Vector2 winnerSize = m_CountdownFont.MeasureString("COMPLETE!");
        Vector2 vecDest;
        vecDest.X = Program.controller.windowSize.Width / 2 - winnerSize.X / 2;
        vecDest.Y = Program.controller.windowSize.Height / 2 - winnerSize.Y / 2;
        spriteHandler.DrawString(m_CountdownFont, "COMPLETE!", vecDest, Color.Red);
      }

      // if PAUSED
      else if (m_playState == PlayState.Paused) {
        // draw "Paused" text
        Vector2 pausedSize = m_CountdownFont.MeasureString("Paused");
        Vector2 dest;
        dest.X = Program.controller.windowSize.Width / 2 - pausedSize.X / 2;
        dest.Y = Program.controller.windowSize.Height / 2 - pausedSize.Y / 2;
        spriteHandler.DrawString(m_CountdownFont, "Paused", dest, Color.Red);

      }



      // We are done, draw everything on screen with help of the end method.
      spriteHandler.End();
    }


    protected void drawActorOnCourse(SpriteBatch spriteHandler, Actor curActor, Vector3 cameraPos) {
      // calculate Z distance from camera (dist down course; this is LOGICAL distance, in the game, not drawing dist)
      int zDistFromPlayer = (int)(curActor.pos.Z - cameraPos.Z);

      // if we need to draw this course object, continue
      //   (we don't need to draw objects that are past the camera or too far to be seen)
      if ((zDistFromPlayer >= 0) && (zDistFromPlayer < MAX_HORIZON_VIEW_DISTANCE)) {

        // calculate the X "offset" of this course object from the camera
        int offset;
        offset = (int)(curActor.pos.X + curActor.objectSize.X/2 - cameraPos.X);

        // find the "wrapping offset" (what our offset would be if we accounted for course wrapping)
        int woffset;
        if (offset > 0) {  // calculate wrapping in the correct direction
          woffset = offset - COURSE_WIDTH;
        } else {
          woffset = offset + COURSE_WIDTH;
        }

        // now, if our wrapping offset is less absolute distance than our normal offset,
        if (Math.Abs(woffset) < Math.Abs(offset)) {
          // use wrapping offset instead
          offset = woffset;
        }

        // will store our destination (on screen)
        Rectangle dest;

        // find the angle of the angle formed by drawing two lines from our "eye" to the object to draw
        //   one line from us to the "top", one from us to the "bottom"
        //   
        //   SIDE VIEW:
        //             ------[ ]
        //         ----      [ ]
        //   (*)---  |angle  [ ]
        //         ----      [ ]
        //             ------[ ]
        //   eye             obj
        //
        // This angle will depend on the DISTANCE we've calculated that we are away from the player "logically"
        double angle = Math.Atan((double)(Program.controller.windowSize.Height) / zDistFromPlayer);

        // now, if that angle is 0, that would mean that the object is too small to be seen
        // if the angle was 90, that would mean the object fills our screen entirely
        // so, here we find the percentage of our screen that the object fills
        double percentage = angle / (Math.PI / 2);
        

        // ********************************
        // calculate object size (to draw):

        // first, if we are not running at 1024x768, adjust the size of our object
        dest.Width = (int)(curActor.objectSize.X * (float)(Program.controller.windowSize.Width) / Program.NATIVE_WIDTH);
        dest.Height = (int)(curActor.objectSize.Y * (float)(Program.controller.windowSize.Height) / Program.NATIVE_HEIGHT);

        // now, scale the object down by the percentage of the screen we should take up
        dest.Width = (int)(dest.Width * percentage);
        dest.Height = (int)(dest.Height * percentage);
        // ********************************


        // now, position our object
        //
        // x position places the center of our course object offset distance from the middle of the screen
        // however, we scale the offset to make distances smaller "farther" away
        dest.X = (int)((double)Program.controller.windowSize.Width / 2 + (percentage * offset) - dest.Width / 2);
        
        //
        // position 5/8 of the object's pixels above the horizon line
        //dest.Y = horizonYPixelPos - dest.Height * 5 / 8;
        double bottom = horizonYPixelPos + (Program.controller.windowSize.Height - horizonYPixelPos) * percentage;
        dest.Y = (int)(bottom - dest.Height + ( cameraPos.Y - curActor.pos.Y )* percentage);





        //double theta = Math.Atan((double)cameraPos.Y / MAX_HORIZON_VIEW_DISTANCE);
        //int distFromHorizon = (int)(MAX_HORIZON_VIEW_DISTANCE - zDistFromPlayer);
        //double y = (double)(Program.controller.windowSize.Height-horizonYPixelPos) * distFromHorizon / MAX_HORIZON_VIEW_DISTANCE;
        //y = 1 / y;
        //double de = cameraPos.Y * Math.Tan(theta);
        //double yp = (de + zDistFromPlayer) * Math.Sin(theta);
        //int bottom = horizonYPixelPos + (int)y;
        //dest.X = Program.controller.windowSize.Width / 2;
        //int bottom = (int)(horizonYPixelPos + (Program.controller.windowSize.Height - horizonYPixelPos) * (percentage));
        //bottom += (int)(cameraPos.Y * percentage);
        //dest.Y = bottom - dest.Height + (int)(cameraPos.Y * percentage);
        //dest.Y = bottom; 
        





        // now finally draw
        curActor.draw(spriteHandler, dest);
      }
    }
    #endregion




    #region Activate()
    public override void Activate() {
      m_playState = PlayState.Beginning;
      m_stateTimer = m_stateTimer.Add(m_stateTimer.Negate());

      // reset our game timer to zero
      m_playTimer = m_playTimer.Subtract(m_playTimer);

      // reset our position, speed, and angle
      m_player.pos.X = COURSE_WIDTH - m_player.objectSize.X / 2;
      m_player.pos.Y = 0;
      m_player.pos.Z = 0;
      m_player.speed = PLAYER_INITIAL_SPEED;
      m_player.vSpeed = 0;
      m_player.angle = 0;
      m_player.state = Player.STATE_MOVING;

      m_player.health = 100;

      moveCameraWithPlayer();

      if (Program.controller.inputToUse != ControllerEngine.InputType.Mouse) {
        // hide the mouse when playing (we show it again when done playing)
        Program.controller.IsMouseVisible = false;
      }
    }
    #endregion

    


    #region Deactivate()
    public override void Deactivate() {
      // show the mouse once again (in case we hid it in Activate())
      Program.controller.IsMouseVisible = true;
      if (songPlaying != null) {
        songPlaying.Stop(AudioStopOptions.Immediate);
      }
      Program.controller.camera.deactivate();
    }
    #endregion




    #region Course Generation
    protected bool generateCourseActors(ContentManager c) {
      bool courseInitSuccess;
      m_staticCourseObjects = new CourseSP();
      courseInitSuccess = m_staticCourseObjects.initialize(COURSE_MAX_Z);

      if (courseInitSuccess == false) return false;

      // this populates our course with some stuff
      for (int i = 0; i < numStaticActorsToGenerate; i++) {
        // randomly choose an object type
        int choice = Program.controller.randomGenerator.Next(21);
        StaticActor t;
        if ((choice >= 0) && (choice < 9)) {
          t = new Tree1();
        } else if ((choice >= 9) && (choice < 13)) {
          t = new Bump1();
        } else if ((choice >= 13) && (choice < 15)) {
          t = new Bump2();
        } else if ((choice >= 15) && (choice < 16)) {
          t = new Rock1();
        } else if ((choice >= 16) && (choice < 19)) {
          t = new Tree2();
        } else {
          t = new Bump3();
        }

        courseInitSuccess = courseInitSuccess && t.initialize(c);
        t.pos.X = Program.controller.randomGenerator.Next(COURSE_WIDTH);
        t.pos.Y = 0;
        // we don't allow any objects from 0-2999 so that you can't get in a position where you can't avoid hitting one
        t.pos.Z = Program.controller.randomGenerator.Next(COURSE_MAX_Z - 3000) + 3000;

        m_staticCourseObjects.add(t);
      }

      // now, generate our snowmen
      for (int i = 1; i <= 100 / Snowman.PERCENT_COURSE_PER_SNOWMAN; i++) {
        // generate a number of them at the same Z position across the course horizontally.
        for (int posX = 0; posX < COURSE_WIDTH; posX += SNOWMAN_X_DIST_APART){
          Snowman s = new Snowman();
          courseInitSuccess = courseInitSuccess && s.initialize(c);
          s.pos.X = posX;
          s.pos.Y = 0;
          s.pos.Z = (float)(i * Snowman.PERCENT_COURSE_PER_SNOWMAN)/100 * COURSE_MAX_Z - 100;
          
          m_staticCourseObjects.add(s);
        }
      }

      // when done populating, we have to do this sort so that objects are drawn in the proper order
      m_staticCourseObjects.sortFarthestToNearest();

      return courseInitSuccess;
    }

    public bool regenerateCourse() {
      ContentManager c = Program.controller.Content;
      return generateCourseActors(c);
    }
    #endregion




  }
}
